package example.localplugin;

import gridsim.Gridlet;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import schedframe.resources.PowerState;
import schedframe.resources.units.ResourceUnit;
import schedframe.scheduling.TaskInterface;
import schedframe.scheduling.events.SchedulingEvent;
import schedframe.scheduling.plugin.grid.ModuleList;
import schedframe.scheduling.utils.ResourceParameterName;
import test.rewolucja.GSSIMJobInterface;
import test.rewolucja.resources.ProcessingElements;
import test.rewolucja.resources.ResourceStatus;
import test.rewolucja.resources.manager.implementation.ClusterResourceManager;
import test.rewolucja.resources.manager.interfaces.ResourceManagerInterface;
import test.rewolucja.resources.physical.base.ComputingResource;
import test.rewolucja.resources.physical.implementation.ComputingNode;
import test.rewolucja.scheduling.JobRegistryInterface;
import test.rewolucja.scheduling.plan.SchedulingPlanInterfaceNew;
import test.rewolucja.scheduling.plan.SchedulingPlanNew;
import test.rewolucja.scheduling.queue.Queue;
import test.rewolucja.scheduling.queue.QueueList;

public class FCFSNodePowerManagementClusterLocalPlugin extends BaseLocalPlugin {

	public FCFSNodePowerManagementClusterLocalPlugin () {
	}

	public SchedulingPlanInterfaceNew schedule(SchedulingEvent event, QueueList queues, JobRegistryInterface jobRegistry,
			ResourceManagerInterface resManager, ModuleList modules) {

		ClusterResourceManager resourceManager = (ClusterResourceManager) resManager;
		SchedulingPlanNew plan = new SchedulingPlanNew();
		// chose the events types to serve.
		// Different actions for different events are possible.
		switch (event.getType()) {
		case START_TASK_EXECUTION:
		case TASK_FINISHED:
			// our tasks are placed only in first queue (see
			// BaseLocalPlugin.placeTasksInQueues() method)
			Queue q = queues.get(0);
			// check all tasks in queue

			for (int i = 0; i < q.size(); i++) {
				GSSIMJobInterface<?> job = q.get(i);
				TaskInterface<?> task = (TaskInterface<?>) job;
				// if status of the tasks in READY
				if (task.getStatus() == Gridlet.READY) {

					Map<ResourceParameterName, ResourceUnit> choosenResources = chooseResourcesForExecution(resourceManager, task);
					if (choosenResources  != null) {
						addToSchedulingPlan(plan, task, choosenResources);
					} else {
						if(harnessNodestoWork(task, resourceManager.getComputingNodes()))
							i--;
					}
				}
			}

			turnOffIdleNodes(resourceManager.getComputingNodes());

			break;
		}
		return plan;
	}
	
	private HashMap<ResourceParameterName, ResourceUnit> chooseResourcesForExecution(ClusterResourceManager resourceManager, TaskInterface<?> task) {

		List<ComputingNode> nodes = resourceManager.getComputingNodes();
		nodes = findSuitableNodes(task, nodes);
		Collections.sort(nodes, new Comparator<ComputingNode>(){
		    public int compare(ComputingNode node1, ComputingNode node2){    
		        return node1.getCategory().getName().compareTo(node2.getCategory().getName());
		    }
		});
		if(nodes.size() > 0)
		{
			HashMap<ResourceParameterName, ResourceUnit> map = new HashMap<ResourceParameterName, ResourceUnit>();
			List<ComputingResource> choosenResources =  new ArrayList<ComputingResource>();
			int cpuRequest;
			try {
				cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
			} catch (NoSuchFieldException e) {
				cpuRequest = 1;
			}
			for (int i = 0; i < nodes.get(0).getProcessors().size() && cpuRequest > 0; i++) {
				if (nodes.get(0).getProcessors().get(i).getStatus() == ResourceStatus.FREE) {
					choosenResources.add(nodes.get(0).getProcessors().get(i));
					cpuRequest--;
				}
			}
			ProcessingElements result = new ProcessingElements(nodes.get(0).getName());
			result.addAll(choosenResources);
			map.put(ResourceParameterName.PROCESSINGELEMENTS, result);
			return map;
		} else 
			return null;
	}
	
	private List<ComputingNode> findSuitableNodes(TaskInterface<?> task, List<ComputingNode> nodes){
		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 1;
		}
		List<ComputingNode> avNodes = new ArrayList<ComputingNode>();
		for(ComputingNode node: nodes){
			if(node.getFreeProcessorsNumber() >= cpuRequest){
				avNodes.add(node);
			}
		}
		return avNodes;
	}

	private void turnOffIdleNodes(List<ComputingNode> nodes){
		for(ComputingNode node : nodes){
			if(node.getFreeProcessorsNumber() == node.getProcessorsNumber()){
				node.getPowerInterface().setPowerState(PowerState.OFF);
			}
		}
	}
	
	private boolean harnessNodestoWork(TaskInterface<?> task, List<ComputingNode> nodes){
		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 1;
		}
		Collections.sort(nodes, new Comparator<ComputingNode>(){
		    public int compare(ComputingNode node1, ComputingNode node2){    
		        return node1.getCategory().getName().compareTo(node2.getCategory().getName());
		    }
		});
		for (int i = 0; i < nodes.size() && cpuRequest > 0; i++) {
			ComputingNode node = nodes.get(i);
			if(node.getPowerInterface().getPowerState() == PowerState.OFF){
				node.getPowerInterface().setPowerState(PowerState.ON);
				cpuRequest -= node.getProcessorsNumber();
			}
		}
		return cpuRequest > 0 ? false : true;
	}
	
	public String getName() {
		return getClass().getName();
	}

	public void init(Properties properties) {
		// no extra initialization is expected.
	}

}